/*
 * btcp_timer.c
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986, 1988, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)tcp_timer.c	8.1 (Berkeley) 6/10/93
 *
 * Modified for x-kernel v3.3
 * Modifications Copyright (c) 1996,1991  Arizona Board of Regents
 *
 * $Revision: 1.3 $
 * $Date: 1996/06/14 21:28:21 $
 */

#include "xkernel.h"
#include "tcp_internal.h"
#include "tcp_fsm.h"
#include "tcp_seq.h"
#include "tcp_timer.h"
#include "tcp_var.h"
#include "tcpip.h"
#include "btcp.h"

static int tcpnodelack = 0;
static int  mytrace = 0;

static int tcp_keepidle = TCPTV_KEEP_IDLE;
	/* LSB: Changed next one for simulator! */
static int tcp_keepintvl = 6 /* TCPTV_KEEPINTVL */;
/* static int tcp_maxidle; */

/*
 * Fast timeout routine for processing delayed acks
 */
void
btcp_fasttimo( ev, arg)
Event ev;
VOID  *arg;
{
    register struct inpcb *inp;
    register struct tcpcb *tp;
    PSTATE *ps=(PSTATE *)arg;
    /* 
     * The original code protected this routine with splnet().  If the timeout
     * thread blocks, there could be a problem.
     */
/*    int s = splnet(); */

    DO_TRACE(ps, TCP_EVENT_FAST_TO, tcpGetTime(), 0, 0);
    inp = ps->tcb.inp_next;
    if (inp)
    for (; inp != &ps->tcb; inp = inp->inp_next)
    	if ((tp = (struct tcpcb *)inp->inp_ppcb) && (tp->t_flags & TF_DELACK)) {
	    tp->t_flags &= ~TF_DELACK;
	    tp->t_flags |= TF_ACKNOW;
	    ps->tcpstat.tcps_delack++;
	    /* What if this blocks??? */
	    (void) (ps->tcp_output)(tp);
    	}
/*    splx(s); */
    evDetach(evSchedule(btcp_fasttimo, (void *)ps, TCP_FAST_INTERVAL));
}

/*
 * Tcp protocol timeout routine called every 500 ms.  Updates the timers in all
 * active tcb's and causes finite state machine actions if timers expire.
 */
void
btcp_slowtimo(ev, arg)
Event ev;
VOID  *arg;
{
    register struct inpcb *ip, *ipnxt;
    register struct tcpcb *tp;
    PSTATE *ps=(PSTATE *)arg;
    /* 
     * The original code protected this routine with splnet().  If the timeout
     * thread blocks, there could be a problem.
     */
/*     int s = splnet(); */
    register int i;

    DO_TRACE(ps, TCP_EVENT_SLOW_TO, tcpGetTime(), 0, 0);
    ps->tcp_maxidle = TCPTV_KEEPCNT * tcp_keepintvl;

    /* Search through tcb's and update active timers. */
    ip = ps->tcb.inp_next;
    if (ip != 0) {
        for (; ip != &ps->tcb; ip = ipnxt) {
	    ipnxt = ip->inp_next;
	    tp = intotcpcb(ip);
	    if (tp == 0)
	    	continue;
	    for (i = 0; i < TCPT_NTIMERS; i++) {
	    	if (tp->t_timer[i] && --tp->t_timer[i] == 0) {
	    	    /* What if this blocks??? */
	    	    (void)btcp_usrreq(tp->t_inpcb->inp_session, PRU_SLOWTIMO,
				      (Msg *)0, (Msg *)i);
	    	    if (ipnxt->inp_prev != ip)
	    	    	goto tpgone;
	    	}
	    }
	    tp->t_idle++;
	    if (tp->t_rtt)
	    	tp->t_rtt++;
tpgone:
	    ;
        }
        ps->tcp_iss += TCP_ISSINCR/PR_SLOWHZ;		/* increment iss */
#ifdef TCP_COMPAT_42
        if ((int)ps->tcp_iss < 0)
            ps->tcp_iss = 0;				/* XXX */
#endif
        ps->tcp_now++;					/* for timestamps */
    }
/*    splx(s); */
    evDetach(evSchedule(btcp_slowtimo, (void *)ps, TCP_SLOW_INTERVAL));
}

#ifndef TUBA_INCLUDE

/*
 * Cancel all timers for TCP tp.
 */
void
btcp_canceltimers(tp)
struct tcpcb *tp;
{
    register int i;

    for (i = 0; i < TCPT_NTIMERS; i++)
    	tp->t_timer[i] = 0;
}

#ifndef XNETSIM
int tcp_backoff[TCP_MAXRXTSHIFT + 1] =
    { 1, 2, 4, 8, 16, 32, 64, 64, 64, 64, 64, 64, 64 };
#endif

/*
 * TCP timer processing.
 */
struct tcpcb *
btcp_timers(tp, timer, arg)
register struct tcpcb *tp;
int             timer;
void            *arg;
{
    register int rexmt;
    PSTATE *ps=(PSTATE *)arg;
    Protl   self=tp->t_inpcb->inp_session->myprotl;
    struct tcpstate *tcpst;

    tcpst = tcpcbtoso(tp)->state;
    DO_TRACE(ps, TCP_EVENT_TIMER, tcpGetTime(), timer, tcpst->tid);
    switch (timer) {

	/*
	 * 2 MSL timeout in shutdown went off.  If we're closed but still
	 * waiting for peer to close and connection has been idle too long, or
	 * if 2MSL time is up from TIME_WAIT, delete connection control block.
	 * Otherwise, check again in a bit.
	 */
	case TCPT_2MSL:
	    if (tp->t_state != TCPS_TIME_WAIT && tp->t_idle <= ps->tcp_maxidle)
	        tp->t_timer[TCPT_2MSL] = tcp_keepintvl/2; /* LSB added /2 */
	    else
	        tp = btcp_destroy(tp);
	    break;

	/*
	 * Retransmission timer went off.  Message has not been acked within
	 * retransmit interval.  Back off to a longer retransmit interval and
	 * retransmit one segment.
	 */
	case TCPT_REXMT:
	    xTrace1(tcpp, 4, "btcp: retransmission timeout (%d)",
	    	    tp->t_rxtshift);
	    if (tp->v_other14 < tp->t_rxtcur)
	    	tp->v_other14 = tp->t_rxtcur;
	    if (++tp->t_rxtshift > TCP_MAXRXTSHIFT) {
	        xTrace0(tcpp, 4, "btcp: too many rexmits, dropping");
		tp->t_rxtshift = TCP_MAXRXTSHIFT;
		ps->tcpstat.tcps_timeoutdrop++;
		tp = btcp_drop(tp, tp->t_softerror ?
			       tp->t_softerror : ETIMEDOUT);
		break;
	    }
	    ps->tcpstat.tcps_rexmttimeo++;
	    /* LSB: prev vers do TCP_REXMTVAL(tp) >> 1 ! */
#ifdef FIX_2
            rexmt = (((tp->t_srtt >> 3) + tp->t_rttvar) >> 2) *
		    tcp_backoff[tp->t_rxtshift];
#else
   	    rexmt = TCP_REXMTVAL(tp) * tcp_backoff[tp->t_rxtshift];   
#endif
	    TCPT_RANGESET(tp->t_rxtcur, rexmt, tp->t_rttmin, TCPTV_REXMTMAX);
	    DO_TRACE(ps, TCP_EVENT_OTHER9, tp->t_rxtcur, 4, tcpst->tid);
	    tp->t_timer[TCPT_REXMT] = tp->t_rxtcur;
	    /*
	     * If losing, let the lower level know and try for a better route.
	     * Also, if we backed off this far, our srtt estimate is probably
	     * bogus.  Clobber it so we'll take the next rtt measurement as our
	     * srtt; move the current srtt into rttvar to keep the current
	     * retransmit times until then.
	     */
	    if (tp->t_rxtshift > TCP_MAXRXTSHIFT / 4) {
/* 	    	in_losing(tp->t_inpcb); */
	    	tp->t_rttvar += (tp->t_srtt >> TCP_RTT_SHIFT);
	    	tp->t_srtt = 0;
	    }
	    tp->snd_nxt = tp->snd_una;
	    /*
	     * If timing a segment in this window, stop the timer.
	     */
	    tp->t_rtt = 0;
	    /*
	     * Close the congestion window down to one segment (we'll open it
	     * by one segment for each ack we get).  Since we probably have a
	     * window's worth of unacked data accumulated, this "slow start"
	     * keeps us from dumping all that data as back-to-back packets
	     * (which might overwhelm an intermediate gateway).
	     *
	     * There are two phases to the opening: Initially we open by one
	     * mss on each ack.  This makes the window size increase
	     * exponentially with time.  If the window is larger than the path
	     * can handle, this exponential growth results in dropped packet(s)
	     * almost immediately.  To get more time between drops but still
	     * "push" the network to take advantage of improving conditions, we
	     * switch from exponential to linear window opening at some
	     * threshhold size.  For a threshhold, we use half the current
	     * window size, truncated to a multiple of the mss.
	     *
	     * (the minimum cwnd that will give us exponential growth is 2 mss.
	     * We don't allow the threshhold to go below this.)
	     */
	    {
	        u_int win = MIN(tp->snd_wnd, tp->snd_cwnd) / 2 / tp->t_maxseg;

		if (win < 2)
		    win = 2;
		tp->snd_cwnd = tp->t_slowstart * tp->t_maxseg;
		tp->snd_ssthresh = win * tp->t_maxseg;
		tp->t_dupacks = 0;
	    }
	    (void) (ps->tcp_output)(tp);
	    break;

	/*
	 * Persistance timer into zero window.  Force a byte to be output,
	 * if possible.
	 */
	case TCPT_PERSIST:
	    ps->tcpstat.tcps_persisttimeo++;
	    tcp_setpersist(tp);
	    tp->t_force = 1;
	    (void) (ps->tcp_output)(tp);
	    tp->t_force = 0;
	    break;

	/*
	 * Keep-alive timer went off; send something or drop connection if
	 * idle for too long.
	 */
	case TCPT_KEEP:
	    ps->tcpstat.tcps_keeptimeo++;
	    if (tp->t_state < TCPS_ESTABLISHED)
	    	goto dropit;
/* LSB ? */
#define SHOULDDOKEEPS(X) 0
	    if (SHOULDDOKEEPS(tp->t_inpcb->inp_session)  &&
	        tp->t_state <= TCPS_CLOSE_WAIT) {
	        if (tp->t_idle >= tcp_keepidle + ps->tcp_maxidle)
		    goto dropit;
		    /*
		     * Send a packet designed to force a response if the peer
		     * is up and reachable: either an ACK if the connection is
		     * still alive, or an RST if the peer has closed the
		     * connection due to timeout or reboot.  Using sequence
		     * number tp->snd_una-1 causes the transmitted zero-length
		     * segment to lie outside the receive window; by the
		     * protocol spec, this requires the correspondent TCP to
		     * respond.
		     */
		    ps->tcpstat.tcps_keepprobe++;
#ifdef TCP_COMPAT_42
		    /*
		     * The keepalive packet must have nonzero length to get a
		     * 4.2 host to respond.
		     */
		    btcp_respond(self, tp, &tp->t_template->ti_t, 
		    	         &tp->t_template->ti_p,
				 tp->rcv_nxt - 1, tp->snd_una - 1, 0);
#else
		    btcp_respond(self, tp, &tp->t_template->ti_t,
				 &tp->t_template->ti_p,
				 tp->rcv_nxt, tp->snd_una - 1, 0);
#endif
		    tp->t_timer[TCPT_KEEP] = tcp_keepintvl;
		}
		else
		    tp->t_timer[TCPT_KEEP] = tcp_keepidle;
		break;
dropit:
		ps->tcpstat.tcps_keepdrops++;
		tp = btcp_drop(tp, ETIMEDOUT);
		break;
	}
	return tp;
}

#endif /* TUBA_INCLUDE */
